@@ -1,14 +1,18 @@
 #[cfg(test)]
 mod test;
 
+use bson::Document;
+
 use crate::{
     bson::doc,
-    cmap::{Command, CommandResponse, StreamDescription},
+    cmap::{Command, StreamDescription},
     error::Result,
     operation::{append_options, Operation, WriteConcernOnlyBody},
     options::{DropDatabaseOptions, WriteConcern},
 };
 
+use super::CommandResponse;
+
 #[derive(Debug)]
 pub(crate) struct DropDatabase {
     target_db: String,
@@ -28,9 +32,12 @@ impl DropDatabase {
 
 impl Operation for DropDatabase {
     type O = ();
+    type Command = Document;
+    type Response = CommandResponse<WriteConcernOnlyBody>;
+
     const NAME: &'static str = "dropDatabase";
 
-    fn build(&self, _description: &StreamDescription) -> Result<Command> {
+    fn build(&mut self, _description: &StreamDescription) -> Result<Command> {
         let mut body = doc! {
             Self::NAME: 1,
         };
@@ -44,8 +51,12 @@ impl Operation for DropDatabase {
         ))
     }
 
-    fn handle_response(&self, response: CommandResponse) -> Result<Self::O> {
-        response.body::<WriteConcernOnlyBody>()?.validate()
+    fn handle_response(
+        &self,
+        response: WriteConcernOnlyBody,
+        _description: &StreamDescription,
+    ) -> Result<Self::O> {
+        response.validate()
     }
 
     fn write_concern(&self) -> Option<&WriteConcern> {
