@@ -1,6 +1,6 @@
 pub(crate) mod server_selection;
 #[cfg(test)]
-mod test;
+pub(crate) mod test;
 
 use std::{
     collections::{HashMap, HashSet},
@@ -13,68 +13,98 @@ use crate::{
     bson::oid::ObjectId,
     client::ClusterTime,
     cmap::Command,
-    error::{ErrorKind, Result},
-    options::{ClientOptions, StreamAddress},
+    options::{ClientOptions, ServerAddress},
     sdam::description::server::{ServerDescription, ServerType},
     selection_criteria::{ReadPreference, SelectionCriteria},
 };
 
 const DEFAULT_HEARTBEAT_FREQUENCY: Duration = Duration::from_secs(10);
 
-/// The TopologyType type, as described by the SDAM spec.
+/// The possible types for a topology.
 #[derive(Debug, Clone, Copy, Eq, PartialEq, Deserialize)]
-pub(crate) enum TopologyType {
+#[non_exhaustive]
+pub enum TopologyType {
+    /// A single mongod server.
     Single,
+
+    /// A replica set with no primary.
     ReplicaSetNoPrimary,
+
+    /// A replica set with a primary.
     ReplicaSetWithPrimary,
+
+    /// A sharded topology.
     Sharded,
+
+    /// A load balanced topology.
+    LoadBalanced,
+
+    /// A topology whose type is not known.
     Unknown,
 }
 
+#[cfg(test)]
+impl TopologyType {
+    fn as_str(&self) -> &'static str {
+        match self {
+            Self::Single => "Single",
+            Self::ReplicaSetNoPrimary => "ReplicaSetNoPrimary",
+            Self::ReplicaSetWithPrimary => "ReplicaSetWithPrimary",
+            Self::Sharded => "Sharded",
+            Self::LoadBalanced => "LoadBalanced",
+            Self::Unknown => "Unknown",
+        }
+    }
+}
+
 impl Default for TopologyType {
     fn default() -> Self {
         TopologyType::Unknown
     }
 }
 
-/// The TopologyDescription type, as described by the SDAM spec.
+/// A description of the most up-to-date information known about a topology.
 #[derive(Debug, Clone)]
+#[non_exhaustive]
 pub(crate) struct TopologyDescription {
     /// Whether or not the topology was initialized with a single seed.
-    single_seed: bool,
+    pub(crate) single_seed: bool,
 
     /// The current type of the topology.
-    topology_type: TopologyType,
+    pub(crate) topology_type: TopologyType,
 
     /// The replica set name of the topology.
-    set_name: Option<String>,
+    pub(crate) set_name: Option<String>,
 
     /// The highest replica set version the driver has seen by a member of the topology.
-    max_set_version: Option<i32>,
+    pub(crate) max_set_version: Option<i32>,
 
     /// The highest replica set election id the driver has seen by a member of the topology.
-    max_election_id: Option<ObjectId>,
+    pub(crate) max_election_id: Option<ObjectId>,
 
     /// Describes the compatibility issue between the driver and server with regards to the
     /// respective supported wire versions.
-    compatibility_error: Option<String>,
+    pub(crate) compatibility_error: Option<String>,
 
     /// Whether or not this topology supports sessions, and if so, what the logicalSessionTimeout
     /// is for them.
-    session_support_status: SessionSupportStatus,
+    pub(crate) session_support_status: SessionSupportStatus,
+
+    /// Whether or not this topology supports transactions.
+    pub(crate) transaction_support_status: TransactionSupportStatus,
 
     /// The highest reported cluster time by any server in this topology.
-    cluster_time: Option<ClusterTime>,
+    pub(crate) cluster_time: Option<ClusterTime>,
 
     /// The amount of latency beyond that of the suitable server with the minimum latency that is
     /// acceptable for a read operation.
-    local_threshold: Option<Duration>,
+    pub(crate) local_threshold: Option<Duration>,
 
     /// The maximum amount of time to wait before checking a given server by sending an isMaster.
-    heartbeat_freq: Option<Duration>,
+    pub(crate) heartbeat_freq: Option<Duration>,
 
     /// The server descriptions of each member of the topology.
-    servers: HashMap<StreamAddress, ServerDescription>,
+    pub(crate) servers: HashMap<ServerAddress, ServerDescription>,
 }
 
 impl PartialEq for TopologyDescription {
@@ -85,34 +115,11 @@ impl PartialEq for TopologyDescription {
         self.compatibility_error == other.compatibility_error
             && self.servers == other.servers
             && self.topology_type == other.topology_type
-            && self.cluster_time == other.cluster_time
     }
 }
 
 impl TopologyDescription {
-    /// Creates a new TopologyDescription with the set of servers initialized to the addresses
-    /// specified in `hosts` and each other field set to its default value.
-    #[cfg(test)]
-    pub(crate) fn new_from_hosts(hosts: Vec<StreamAddress>) -> Self {
-        Self {
-            single_seed: false,
-            topology_type: TopologyType::Unknown,
-            set_name: None,
-            max_set_version: None,
-            max_election_id: None,
-            compatibility_error: None,
-            session_support_status: Default::default(),
-            cluster_time: None,
-            local_threshold: None,
-            heartbeat_freq: None,
-            servers: hosts
-                .into_iter()
-                .map(|address| (address.clone(), ServerDescription::new(address, None)))
-                .collect(),
-        }
-    }
-
-    pub(crate) fn new(options: ClientOptions) -> Result<Self> {
+    pub(crate) fn new(options: ClientOptions) -> crate::error::Result<Self> {
         verify_max_staleness(
             options
                 .selection_criteria
@@ -124,6 +131,8 @@ impl TopologyDescription {
             TopologyType::Single
         } else if options.repl_set_name.is_some() {
             TopologyType::ReplicaSetNoPrimary
+        } else if options.load_balanced.unwrap_or(false) {
+            TopologyType::LoadBalanced
         } else {
             TopologyType::Unknown
         };
@@ -132,12 +141,24 @@ impl TopologyDescription {
             .hosts
             .into_iter()
             .map(|address| {
-                let description = ServerDescription::new(address.clone(), None);
+                let description = if topology_type == TopologyType::LoadBalanced {
+                    ServerDescription::new_load_balancer(address.clone())
+                } else {
+                    ServerDescription::new(address.clone(), None)
+                };
 
                 (address, description)
             })
             .collect();
 
+        let session_support_status = if topology_type == TopologyType::LoadBalanced {
+            SessionSupportStatus::Supported {
+                logical_session_timeout: None,
+            }
+        } else {
+            SessionSupportStatus::Undetermined
+        };
+
         Ok(Self {
             single_seed: servers.len() == 1,
             topology_type,
@@ -145,7 +166,8 @@ impl TopologyDescription {
             max_set_version: None,
             max_election_id: None,
             compatibility_error: None,
-            session_support_status: SessionSupportStatus::Undetermined,
+            session_support_status,
+            transaction_support_status: TransactionSupportStatus::Undetermined,
             cluster_time: None,
             local_threshold: options.local_threshold,
             heartbeat_freq: options.heartbeat_freq,
@@ -153,12 +175,29 @@ impl TopologyDescription {
         })
     }
 
+    pub(crate) fn new_empty() -> Self {
+        Self {
+            single_seed: false,
+            topology_type: TopologyType::Unknown,
+            set_name: None,
+            max_set_version: None,
+            max_election_id: None,
+            compatibility_error: None,
+            session_support_status: SessionSupportStatus::Undetermined,
+            transaction_support_status: TransactionSupportStatus::Undetermined,
+            cluster_time: None,
+            local_threshold: None,
+            heartbeat_freq: None,
+            servers: HashMap::new(),
+        }
+    }
+
     /// Gets the topology type of the cluster.
     pub(crate) fn topology_type(&self) -> TopologyType {
         self.topology_type
     }
 
-    pub(crate) fn server_addresses(&self) -> impl Iterator<Item = &StreamAddress> {
+    pub(crate) fn server_addresses(&self) -> impl Iterator<Item = &ServerAddress> {
         self.servers.keys()
     }
 
@@ -168,21 +207,22 @@ impl TopologyDescription {
 
     pub(crate) fn get_server_description(
         &self,
-        address: &StreamAddress,
+        address: &ServerAddress,
     ) -> Option<&ServerDescription> {
         self.servers.get(address)
     }
 
-    pub(crate) fn update_command_with_read_pref(
+    pub(crate) fn update_command_with_read_pref<T>(
         &self,
         server_type: ServerType,
-        command: &mut Command,
+        command: &mut Command<T>,
         criteria: Option<&SelectionCriteria>,
     ) {
         match (self.topology_type, server_type) {
             (TopologyType::Sharded, ServerType::Mongos)
-            | (TopologyType::Single, ServerType::Mongos) => {
-                self.update_command_read_pref_for_mongos(command, criteria);
+            | (TopologyType::Single, ServerType::Mongos)
+            | (TopologyType::LoadBalanced, _) => {
+                self.update_command_read_pref_for_mongos(command, criteria)
             }
             (TopologyType::Single, ServerType::Standalone) => {}
             (TopologyType::Single, _) => {
@@ -197,7 +237,7 @@ impl TopologyDescription {
                     Some(other) => other,
                 };
 
-                command.set_read_preference(resolved_read_pref);
+                command.set_read_preference(resolved_read_pref)
             }
             _ => {
                 let read_pref = match criteria {
@@ -207,40 +247,30 @@ impl TopologyDescription {
                     },
                     None => ReadPreference::Primary,
                 };
-                command.set_read_preference(read_pref);
+                command.set_read_preference(read_pref)
             }
         }
     }
 
-    fn update_command_read_pref_for_mongos(
+    fn update_command_read_pref_for_mongos<T>(
         &self,
-        command: &mut Command,
+        command: &mut Command<T>,
         criteria: Option<&SelectionCriteria>,
     ) {
-        match criteria {
-            Some(SelectionCriteria::ReadPreference(ReadPreference::Secondary { ref options })) => {
-                command.set_read_preference(ReadPreference::Secondary {
-                    options: options.clone(),
-                });
+        let read_preference = match criteria {
+            Some(SelectionCriteria::ReadPreference(rp)) => rp,
+            _ => return,
+        };
+        match read_preference {
+            ReadPreference::Secondary { .. }
+            | ReadPreference::PrimaryPreferred { .. }
+            | ReadPreference::Nearest { .. } => {
+                command.set_read_preference(read_preference.clone())
             }
-            Some(SelectionCriteria::ReadPreference(ReadPreference::PrimaryPreferred {
-                ref options,
-            })) => {
-                command.set_read_preference(ReadPreference::PrimaryPreferred {
-                    options: options.clone(),
-                });
-            }
-            Some(SelectionCriteria::ReadPreference(ReadPreference::SecondaryPreferred {
-                ref options,
-            })) if options.max_staleness.is_some() || options.tag_sets.is_some() => {
-                command.set_read_preference(ReadPreference::SecondaryPreferred {
-                    options: options.clone(),
-                });
-            }
-            Some(SelectionCriteria::ReadPreference(ReadPreference::Nearest { ref options })) => {
-                command.set_read_preference(ReadPreference::Nearest {
-                    options: options.clone(),
-                });
+            ReadPreference::SecondaryPreferred { ref options }
+                if options.max_staleness.is_some() || options.tag_sets.is_some() =>
+            {
+                command.set_read_preference(read_preference.clone())
             }
             _ => {}
         }
@@ -286,16 +316,11 @@ impl TopologyDescription {
     /// Updates the topology's logical session timeout value based on the server's value for it.
     fn update_session_support_status(&mut self, server_description: &ServerDescription) {
         if !server_description.server_type.is_data_bearing() {
-            return;
-        }
-
-        if server_description.server_type == ServerType::Standalone {
-            self.session_support_status = SessionSupportStatus::Unsupported {
-                logical_session_timeout: server_description
-                    .logical_session_timeout()
-                    .ok()
-                    .flatten(),
-            };
+            if let TopologyType::Single = self.topology_type {
+                self.session_support_status = SessionSupportStatus::Unsupported {
+                    logical_session_timeout: None,
+                };
+            }
             return;
         }
 
@@ -305,12 +330,12 @@ impl TopologyDescription {
                     logical_session_timeout: topology_timeout,
                 } => {
                     self.session_support_status = SessionSupportStatus::Supported {
-                        logical_session_timeout: std::cmp::min(timeout, topology_timeout),
+                        logical_session_timeout: std::cmp::min(Some(timeout), topology_timeout),
                     };
                 }
                 SessionSupportStatus::Undetermined => {
                     self.session_support_status = SessionSupportStatus::Supported {
-                        logical_session_timeout: timeout,
+                        logical_session_timeout: Some(timeout),
                     }
                 }
                 SessionSupportStatus::Unsupported { .. } => {
@@ -327,7 +352,7 @@ impl TopologyDescription {
                     match min_timeout {
                         Some(timeout) => {
                             self.session_support_status = SessionSupportStatus::Supported {
-                                logical_session_timeout: timeout,
+                                logical_session_timeout: Some(timeout),
                             }
                         }
                         None => {
@@ -349,6 +374,26 @@ impl TopologyDescription {
         }
     }
 
+    /// Updates the topology's transaction support status based on its session support status and
+    /// the server description's max wire version.
+    fn update_transaction_support_status(&mut self, server_description: &ServerDescription) {
+        if !matches!(
+            self.session_support_status,
+            SessionSupportStatus::Supported { .. }
+        ) {
+            self.transaction_support_status = TransactionSupportStatus::Unsupported;
+        }
+        if let Ok(Some(max_wire_version)) = server_description.max_wire_version() {
+            self.transaction_support_status = if max_wire_version < 7
+                || (max_wire_version < 8 && self.topology_type == TopologyType::Sharded)
+            {
+                TransactionSupportStatus::Unsupported
+            } else {
+                TransactionSupportStatus::Supported
+            }
+        }
+    }
+
     /// Sets the topology's cluster time to the provided one if it is higher than the currently
     /// recorded one.
     pub(crate) fn advance_cluster_time(&mut self, cluster_time: &ClusterTime) {
@@ -363,28 +408,39 @@ impl TopologyDescription {
     ///
     /// The returned `TopologyDescriptionDiff` refers to the changes reflected in the provided
     /// description. For example, if the provided description has a server in it that this
-    /// description does not, it will be returned in the `new_servers` field.
-    pub(crate) fn diff(&self, other: &TopologyDescription) -> Option<TopologyDescriptionDiff> {
+    /// description does not, it will be returned in the `added_addresses` field.
+    pub(crate) fn diff<'a>(
+        &'a self,
+        other: &'a TopologyDescription,
+    ) -> Option<TopologyDescriptionDiff> {
         if self == other {
             return None;
         }
 
-        let addresses: HashSet<&StreamAddress> = self.server_addresses().collect();
-        let other_addresses: HashSet<&StreamAddress> = other.server_addresses().collect();
+        let addresses: HashSet<&ServerAddress> = self.server_addresses().collect();
+        let other_addresses: HashSet<&ServerAddress> = other.server_addresses().collect();
+
+        let changed_servers = self
+            .servers
+            .iter()
+            .filter_map(|(address, description)| match other.servers.get(address) {
+                Some(other_description) if description != other_description => {
+                    Some((address, (description, other_description)))
+                }
+                _ => None,
+            });
 
         Some(TopologyDescriptionDiff {
-            new_addresses: other_addresses
-                .difference(&addresses)
-                .cloned()
-                .cloned()
-                .collect(),
+            removed_addresses: addresses.difference(&other_addresses).cloned().collect(),
+            added_addresses: other_addresses.difference(&addresses).cloned().collect(),
+            changed_servers: changed_servers.collect(),
         })
     }
 
     /// Syncs the set of servers in the description to those in `hosts`. Servers in the set not
     /// already present in the cluster will be added, and servers in the cluster not present in the
     /// set will be removed.
-    pub(crate) fn sync_hosts(&mut self, hosts: &HashSet<StreamAddress>) {
+    pub(crate) fn sync_hosts(&mut self, hosts: &HashSet<ServerAddress>) {
         self.add_new_servers_from_addresses(hosts.iter());
         self.servers.retain(|host, _| hosts.contains(host));
     }
@@ -393,9 +449,16 @@ impl TopologyDescription {
         self.session_support_status
     }
 
+    pub(crate) fn transaction_support_status(&self) -> TransactionSupportStatus {
+        self.transaction_support_status
+    }
+
     /// Update the topology based on the new information about the topology contained by the
     /// ServerDescription.
-    pub(crate) fn update(&mut self, mut server_description: ServerDescription) -> Result<()> {
+    pub(crate) fn update(
+        &mut self,
+        mut server_description: ServerDescription,
+    ) -> Result<(), String> {
         // Ignore updates from servers not currently in the cluster.
         if !self.servers.contains_key(&server_description.address) {
             return Ok(());
@@ -414,6 +477,9 @@ impl TopologyDescription {
         // Update the topology's min logicalSessionTimeout.
         self.update_session_support_status(&server_description);
 
+        // Update the topology's transaction support status.
+        self.update_transaction_support_status(&server_description);
+
         // Update the topology's max reported $clusterTime.
         if let Some(ref cluster_time) = server_description.cluster_time().ok().flatten() {
             self.advance_cluster_time(cluster_time);
@@ -421,7 +487,7 @@ impl TopologyDescription {
 
         // Update the topology description based on the current topology type.
         match self.topology_type {
-            TopologyType::Single => {}
+            TopologyType::Single | TopologyType::LoadBalanced => {}
             TopologyType::Unknown => self.update_unknown_topology(server_description)?,
             TopologyType::Sharded => self.update_sharded_topology(server_description),
             TopologyType::ReplicaSetNoPrimary => {
@@ -439,21 +505,27 @@ impl TopologyDescription {
     }
 
     /// Update the Unknown topology description based on the server description.
-    fn update_unknown_topology(&mut self, server_description: ServerDescription) -> Result<()> {
+    fn update_unknown_topology(
+        &mut self,
+        server_description: ServerDescription,
+    ) -> Result<(), String> {
         match server_description.server_type {
-            ServerType::Unknown | ServerType::RSGhost => {}
+            ServerType::Unknown | ServerType::RsGhost => {}
             ServerType::Standalone => {
                 self.update_unknown_with_standalone_server(server_description)
             }
             ServerType::Mongos => self.topology_type = TopologyType::Sharded,
-            ServerType::RSPrimary => {
+            ServerType::RsPrimary => {
                 self.topology_type = TopologyType::ReplicaSetWithPrimary;
                 self.update_rs_from_primary_server(server_description)?;
             }
-            ServerType::RSSecondary | ServerType::RSArbiter | ServerType::RSOther => {
+            ServerType::RsSecondary | ServerType::RsArbiter | ServerType::RsOther => {
                 self.topology_type = TopologyType::ReplicaSetNoPrimary;
                 self.update_rs_without_primary_server(server_description)?;
             }
+            ServerType::LoadBalancer => {
+                return Err("cannot transition to a load balancer".to_string())
+            }
         }
 
         Ok(())
@@ -473,19 +545,22 @@ impl TopologyDescription {
     fn update_replica_set_no_primary_topology(
         &mut self,
         server_description: ServerDescription,
-    ) -> Result<()> {
+    ) -> Result<(), String> {
         match server_description.server_type {
-            ServerType::Unknown | ServerType::RSGhost => {}
+            ServerType::Unknown | ServerType::RsGhost => {}
             ServerType::Standalone | ServerType::Mongos => {
                 self.servers.remove(&server_description.address);
             }
-            ServerType::RSPrimary => {
+            ServerType::RsPrimary => {
                 self.topology_type = TopologyType::ReplicaSetWithPrimary;
                 self.update_rs_from_primary_server(server_description)?
             }
-            ServerType::RSSecondary | ServerType::RSArbiter | ServerType::RSOther => {
+            ServerType::RsSecondary | ServerType::RsArbiter | ServerType::RsOther => {
                 self.update_rs_without_primary_server(server_description)?;
             }
+            ServerType::LoadBalancer => {
+                return Err("cannot transition to a load balancer".to_string())
+            }
         }
 
         Ok(())
@@ -495,19 +570,22 @@ impl TopologyDescription {
     fn update_replica_set_with_primary_topology(
         &mut self,
         server_description: ServerDescription,
-    ) -> Result<()> {
+    ) -> Result<(), String> {
         match server_description.server_type {
-            ServerType::Unknown | ServerType::RSGhost => {
+            ServerType::Unknown | ServerType::RsGhost => {
                 self.record_primary_state();
             }
             ServerType::Standalone | ServerType::Mongos => {
                 self.servers.remove(&server_description.address);
                 self.record_primary_state();
             }
-            ServerType::RSPrimary => self.update_rs_from_primary_server(server_description)?,
-            ServerType::RSSecondary | ServerType::RSArbiter | ServerType::RSOther => {
+            ServerType::RsPrimary => self.update_rs_from_primary_server(server_description)?,
+            ServerType::RsSecondary | ServerType::RsArbiter | ServerType::RsOther => {
                 self.update_rs_with_primary_from_member(server_description)?;
             }
+            ServerType::LoadBalancer => {
+                return Err("cannot transition to a load balancer".to_string())
+            }
         }
 
         Ok(())
@@ -527,7 +605,7 @@ impl TopologyDescription {
     fn update_rs_without_primary_server(
         &mut self,
         server_description: ServerDescription,
-    ) -> Result<()> {
+    ) -> Result<(), String> {
         if self.set_name.is_none() {
             self.set_name = server_description.set_name()?;
         } else if self.set_name != server_description.set_name()? {
@@ -550,7 +628,7 @@ impl TopologyDescription {
     fn update_rs_with_primary_from_member(
         &mut self,
         server_description: ServerDescription,
-    ) -> Result<()> {
+    ) -> Result<(), String> {
         if self.set_name != server_description.set_name()? {
             self.servers.remove(&server_description.address);
             self.record_primary_state();
@@ -572,7 +650,7 @@ impl TopologyDescription {
     fn update_rs_from_primary_server(
         &mut self,
         server_description: ServerDescription,
-    ) -> Result<()> {
+    ) -> Result<(), String> {
         if self.set_name.is_none() {
             self.set_name = server_description.set_name()?;
         } else if self.set_name != server_description.set_name()? {
@@ -624,7 +702,7 @@ impl TopologyDescription {
                 continue;
             }
 
-            if let ServerType::RSPrimary = self.servers.get(&address).unwrap().server_type {
+            if let ServerType::RsPrimary = self.servers.get(&address).unwrap().server_type {
                 self.servers
                     .insert(address.clone(), ServerDescription::new(address, None));
             }
@@ -652,7 +730,7 @@ impl TopologyDescription {
         self.topology_type = if self
             .servers
             .values()
-            .any(|server| server.server_type == ServerType::RSPrimary)
+            .any(|server| server.server_type == ServerType::RsPrimary)
         {
             TopologyType::ReplicaSetWithPrimary
         } else {
@@ -661,8 +739,13 @@ impl TopologyDescription {
     }
 
     /// Create a new ServerDescription for each address and add it to the topology.
-    fn add_new_servers<'a>(&mut self, servers: impl Iterator<Item = &'a String>) -> Result<()> {
-        let servers: Result<Vec<_>> = servers.map(|server| StreamAddress::parse(server)).collect();
+    fn add_new_servers<'a>(
+        &mut self,
+        servers: impl Iterator<Item = &'a String>,
+    ) -> Result<(), String> {
+        let servers: Result<Vec<_>, String> = servers
+            .map(|server| ServerAddress::parse(server).map_err(|e| e.to_string()))
+            .collect();
 
         self.add_new_servers_from_addresses(servers?.iter());
         Ok(())
@@ -671,7 +754,7 @@ impl TopologyDescription {
     /// Create a new ServerDescription for each address and add it to the topology.
     fn add_new_servers_from_addresses<'a>(
         &mut self,
-        servers: impl Iterator<Item = &'a StreamAddress>,
+        servers: impl Iterator<Item = &'a ServerAddress>,
     ) {
         for server in servers {
             if !self.servers.contains_key(server) {
@@ -700,7 +783,9 @@ pub(crate) enum SessionSupportStatus {
 
     /// Sessions are supported by this topology. This is the minimum timeout of all data-bearing
     /// servers in the deployment.
-    Supported { logical_session_timeout: Duration },
+    Supported {
+        logical_session_timeout: Option<Duration>,
+    },
 }
 
 impl Default for SessionSupportStatus {
@@ -719,27 +804,57 @@ impl SessionSupportStatus {
             } => *logical_session_timeout,
             Self::Supported {
                 logical_session_timeout,
-            } => Some(*logical_session_timeout),
+            } => *logical_session_timeout,
         }
     }
 }
 
+/// Enum representing whether transactions are supported by the topology.
+#[derive(Debug, Clone, Copy, PartialEq)]
+pub(crate) enum TransactionSupportStatus {
+    /// It is not known yet whether the topology supports transactions. This is possible if no
+    /// data-bearing servers have updated the `TopologyDescription` yet.
+    Undetermined,
+
+    /// Transactions are not supported by this topology.
+    Unsupported,
+
+    /// Transactions are supported by this topology. A topology supports transactions if it
+    /// supports sessions and its maxWireVersion >= 7. If the topology is sharded, maxWireVersion
+    /// must be >= 8 for transactions to be supported.
+    ///
+    /// Note that meeting these conditions does not guarantee that a deployment
+    /// supports transactions; any other missing qualification will be reported by the server.
+    Supported,
+}
+
+impl Default for TransactionSupportStatus {
+    fn default() -> Self {
+        Self::Undetermined
+    }
+}
+
 /// A struct representing the diff between two `TopologyDescription`s.
 /// Returned from `TopologyDescription::diff`.
 #[derive(Debug)]
-pub(crate) struct TopologyDescriptionDiff {
-    pub(crate) new_addresses: HashSet<StreamAddress>,
+pub(crate) struct TopologyDescriptionDiff<'a> {
+    pub(crate) removed_addresses: HashSet<&'a ServerAddress>,
+    pub(crate) added_addresses: HashSet<&'a ServerAddress>,
+    pub(crate) changed_servers:
+        HashMap<&'a ServerAddress, (&'a ServerDescription, &'a ServerDescription)>,
 }
 
-fn verify_max_staleness(max_staleness: Option<Duration>) -> Result<()> {
+fn verify_max_staleness(max_staleness: Option<Duration>) -> crate::error::Result<()> {
+    verify_max_staleness_inner(max_staleness)
+        .map_err(|s| crate::error::ErrorKind::InvalidArgument { message: s }.into())
+}
+
+fn verify_max_staleness_inner(max_staleness: Option<Duration>) -> std::result::Result<(), String> {
     if max_staleness
         .map(|staleness| staleness > Duration::from_secs(0) && staleness < Duration::from_secs(90))
         .unwrap_or(false)
     {
-        return Err(ErrorKind::ArgumentError {
-            message: "max staleness cannot be both positive and below 90 seconds".into(),
-        }
-        .into());
+        return Err("max staleness cannot be both positive and below 90 seconds".into());
     }
 
     Ok(())
